;  USB Switcher.
;
;  Copyright 2013 Akinori Furuta<afuruta@m7.dion.ne.jp>.
;  All rights reserved.
;
;  Redistribution and use in source and binary forms, with or without 
;  modification, are permitted provided that the following conditions 
;  are met:
;
;  1. Redistributions of source code must retain the above copyright notice,
;     this list of conditions and the following disclaimer.
;
;  2. Redistributions in binary form must reproduce the above copyright notice,
;     this list of conditions and the following disclaimer in the documentation
;     and/or other materials provided with the distribution.
;
;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
;  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
;  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
;  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
;  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
;  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	LIST
	PROCESSOR	16F688
	#include	"P16F688.INC"
	#include	"pic12ext.inc"
	__config	  _FCMEN_OFF & _IESO_OFF & _BOD_ON & _CPD_OFF & _CP_OFF & _MCLRE_ON & _PWRTE_ON & _WDT_OFF & _HS_OSC

	__idlocs	0xff02
			;In our house, Sign the ID number as follows.
			;0x00xx - 0xefxx MSB8: chip kind in hex, LSB8:revision in hex.
			;0xf0xx - 0xfexx Reserved.
			;0xffxx Test purpose.

	;_FCMEN_OFF:	Fail-Safe Clock Monitor is disabled
	;_IESO_OFF:	Internal External Switchover mode is disabled
	;_BOD_ON:	Brown-out Detect on.
	;_CPD_OFF:	Data Code Protection off.
	;_CP_OFF:	Code Protection off.
	;_MCLRE_ON:	/MCLR pin as /MasterClear
	;_PWRTE_ON:	Powerup timer Enabled.
	;_WDT_OFF:	Watch dog timer Disabled.
	;_HS_OSC:	High Speed crystal/resonator oscillator.

; OSC:  20.000MHz
; INST:  5.000MHz

OSC_FREQ	EQU	(D'20000000')
INST_FREQ	EQU	(OSC_FREQ / D'4')

EUSART_BAUD	EQU	(D'38400')
EUSART_BRG	EQU	(D'129')	;SYNC=0, BRG16=1, BRGH=1

;
GPIO_RA0_DIN0		EQU	D'0'	;(I) DigitalIn0
GPIO_RA1_DIN1		EQU	D'1'	;(I) DigitalIn1
GPIO_RA2_AIN0		EQU	D'2'	;(I) AnalogIn0
GPIO_RA3_MCLR		EQU	D'3'	;(I) MasterClear.
GPIO_RA4_OSC2		EQU	D'4'	;(I) Ceramic Resonator OSC2.
GPIO_RA5_OSC1		EQU	D'5'	;(I) Ceramic Resonator OSC1.

GPIO_RA0M_DIN0		EQU	(H'1' << GPIO_RA0_DIN0)
GPIO_RA1M_DIN1		EQU	(H'1' << GPIO_RA1_DIN1)
GPIO_RA2M_AIN0		EQU	(H'1' << GPIO_RA2_AIN0)
GPIO_RA3M_MCLR		EQU	(H'1' << GPIO_RA3_MCLR)
GPIO_RA4M_OSC2		EQU	(H'1' << GPIO_RA4_OSC2)
GPIO_RA5M_OSC1		EQU	(H'1' << GPIO_RA5_OSC1)


GPIO_RC0_VBUS		EQU	D'0'	;(O) VBUS switch control. 1: On, 0: Off.
GPIO_RC1_DPM		EQU	D'1'	;(O) D+, D- switch control. 1: On, 0: Off.
GPIO_RC2_EXT0		EQU	D'2'	;(O) External out 0.
GPIO_RC3_EXT1		EQU	D'3'	;(O) External out 1.
GPIO_RC4_TX		EQU	D'4'	;(O) Serial Transmit.
GPIO_RC5_RX		EQU	D'5'	;(I) Serial Receive.

GPIO_RC0M_VBUS		EQU	(H'1' << GPIO_RC0_VBUS)
GPIO_RC1M_DPM		EQU	(H'1' << GPIO_RC1_DPM)
GPIO_RC2M_EXT0		EQU	(H'1' << GPIO_RC2_EXT0)
GPIO_RC3M_EXT1		EQU	(H'1' << GPIO_RC3_EXT1)
GPIO_RC4M_TX		EQU	(H'1' << GPIO_RC4_TX)
GPIO_RC5M_RX		EQU	(H'1' << GPIO_RC5_RX)

ADCON0_M_ADFM		EQU	(H'1' << ADFM) ; 0: Left,     1: Right
ADCON0_M_VFCG		EQU	(H'1' << VCFG) ; 0: Vdd,      1: Vref
ADCON0_M_GO		EQU	(H'1' << GO)   ; 0: Done,     1: Go
ADCON0_M_ADON		EQU	(H'1' << ADON) ; 0: shut-off, 1: operating

ADCON1_CONV_FREQ	EQU	(D'50000')	;50KHz as 20uS

gpr0	UDATA	;@note refer .lkr to see what is gpr0.
RX_CHAR		RES	1
ADCON0_TMP	RES	1
CONV_HEX_L	RES	1
CONV_HEX_H	RES	1

page0	CODE	0x00 ;@note refer .lkr to see what is page0.
RESET
	GOTO	START	;0x00
	NOP		;0x01
	NOP		;0x02
	NOP		;0x03
INTSERVICE		;0x04
	; No interrupt service.
	RETFIE

START
	;Start up.

	;B1.OPTION_REG
	;7:GPPU#=0 en, 6:INTEG=0 nocare, 5:T0CS=0 Internal, 4:T0SE=0 nocare, 
	;3:PSA=0 Use for TMR0, 2..0:PSA2..PSA0=111b 1/128
	MOVLW	0x07
	BANKSEL	OPTION_REG
	MOVWF	OPTION_REG

	;Initialize WatchDog
	;B0.WDTCON
	;7..4:Undefined
	;4..1:WDTPS=9 31K/16384 Hz
	;0:SWDTEN=1 WDTenabled.
	MOVLW	0x13
	BANKSEL	WDTCON
	MOVWF	WDTCON
	CLRWDT

	;B1.PCON
	;5:ULPWUE=0 disabled, 4:SBODEN=0 disabled
	;1:POR#=1 no-por, 0:BOD#=1 no brown-out
	MOVLW	0x03
	BANKSEL	PCON
	MOVWF	PCON

	CLRWDT

	;B1.PIE1
	;Keep Default (i.e. Disable all).

	;Bx.INTCON (Initial)
	;Keep Default (i.e. Disable all).

	;B1.OSCCON
	;Keep Default.

	;B1.OSCTUNE
	;Don't modify.

	;Initialize Comparator.

	;B0.CMCON0
	;Comparator: Off.
	MOVLW	0x07
	BANKSEL	CMCON0
	MOVWF	CMCON0

	;B0.CMCON1
	;1:T1GSS=1, 0:CMSYNC=0
	;Don' touch up CMCON1.

	;B1.VRCON
	;Keep default.

	;Initialize ports.

	;B1.WPU
	;x:WPUx=1 pu, x:WPUx=0 float
	MOVLW	0x07
	BANKSEL	WPU
	MOVWF	WPU

	;B1.TRISA
	;x:TRISIOx=1 in, x:TRISIOx=0 out
	MOVLW	0x3F
	BANKSEL	TRISA
	MOVWF	TRISA

	;B0.PORTA (All input)
	MOVLW	0x00
	BANKSEL	PORTA
	MOVWF	PORTA

	;B1.IOCA
	;x:IOCAx=1 enabled, x:IOCAx=0 disabled
	MOVLW	0x00
	BANKSEL	IOCA
	MOVWF	IOCA

	;B0.PORTC
	MOVLW	GPIO_RC1M_DPM | GPIO_RC0M_VBUS
	BANKSEL	PORTC
	MOVWF	PORTC

	;B1.TRISC
	;x:TRISIOx=1 in, x:TRISIOx=0 out
	;Refer to.
	;10.0 ENHANCED UNIVERSAL ASYNCHRONOUS RECEIVER TRANSMITTER (EUSART)
	MOVLW	GPIO_RC5M_RX | GPIO_RC4M_TX
	BANKSEL	TRISC
	MOVWF	TRISC

	;Initialize EUSART
	;B0.TXSTA
	;b7: CSRC=0 (Don'care), b6: TX9=0 (8bit), b5: TXEN=1 (Enable TX), b4: SYNC=0 (Async).
	;b3: SENDB=0 (Normal Send), b2: BRGH=1 (High speed), b1: TRMT=0 (RO-bit), b0: TX9D=0 (Don't care).
	MOVLW	0x24
	BANKSEL	TXSTA
	MOVWF	TXSTA

	;B0.RCSTA
	;b7: SPEN=1 (Enable), b6: RX9=0 (8bit), b5: SREN=0 (Don't care), b4: CREN=1 (Enable RX),
	;b3: ADDEN=0 (Don't care), b2: FERR=0 (RO-bit), b1: OERR=0 (RO-bit), b0: RX9D=0 (RO-bit).
	MOVLW	0x90
	BANKSEL	RCSTA
	MOVWF	RCSTA

	;B0.BAUDCTL
	;b7: ABDOVF=0 (RO-bit), b6: RCIDL=0 (R0-bit), b5: Unimp=0, b4: SCKP=0 (TX Polarity Pos),
	;b3: BRG16=1 (16bit mode), b2: Unimp=0, b1: WUE=0 (Disable wakeup), b0: ABDEN=0 (Disable Auto Baud).
	MOVLW	0x08
	BANKSEL	BAUDCTL
	MOVWF	BAUDCTL

	;B0.SPBRGH
	MOVLW	( EUSART_BRG >> 8 )
	BANKSEL	SPBRGH
	MOVWF	SPBRGH

	;B0.SPBRG
	MOVLW	( EUSART_BRG & 0xff )
	BANKSEL	SPBRG
	MOVWF	SPBRG

	;Initialize A/D converter.

	;B1.ANSEL
	;bi: ANSi=0 (Digital), ANSi=1 (Analog)
	MOVLW	0x07
	BANKSEL	ANSEL
	MOVWF	ANSEL

	;B1.ADCON1
	;b7: unimp=0, b6..b4: ADCS2..ADCS0=0b010 (32Tosc)
	;b3..b0: unimp=0
	MOVLW	0x20
	BANKSEL	ADCON1
	MOVWF	ADCON1

	;B0.ADCON0
	;7:ADFM=1 right, 6:VCGF=0 vdd, 5..4:unimp=0, 
	;3..2:CHS1..0=00 an0, 1:GO/DONE#=0 compl, 0:ADON=1 operating
	MOVLW	ADCON0_M_ADFM+ADCON0_M_ADON
	BANKSEL	ADCON0
	MOVWF	ADCON0

	;Initialize Timer.

	;B0.TMR0
	;Don' touch up TIMER0.

	;TIMER1
	;B0.T1CON
	;Don' touch up TIMER1.

	;B1.EECON1
	;Not using.

	;B1.EECON2
	;Not using.

	;B1.OSCCAL
	;Don't modify.
	MOVLW	0x00
	MOVWF	RX_CHAR
MAIN_LOOP
	CLRWDT
	BANKSEL	PIR1
	BTFSS	PIR1,RCIF	;Test EUSART
	GOTO	MAIN_LOOP
	;Received
	MOVF	RCREG,W
	MOVWF	RX_CHAR
	BTFSS	PIR1,TXIF	;Wait TX ready.
	GOTO	$-1
	MOVWF	TXREG		;Echo back
	MOVLW	'0'-1
	SUBWF	RX_CHAR,W
	JBE	HEX_AF
	MOVLW	'9'+1
	SUBWF	RX_CHAR,W
	JAE	HEX_AF
	MOVLW	'0'		;Receive '0'..'9'
	SUBWF	RX_CHAR,W
	MOVWF	PORTC		;Out PORT C
	GOTO	PC_OUT_NEXT
HEX_AF
	MOVLW	'a'-1
	SUBWF	RX_CHAR,W
	JBE	PC_OUT_NEXT
	MOVLW	'f'+1
	SUBWF	RX_CHAR,W
	JAE	PC_OUT_NEXT
	MOVLW	'a'-D'10'	;Receive 'a'..'f'
	SUBWF	RX_CHAR,W
	MOVWF	PORTC		;Out PORT C
PC_OUT_NEXT
	BANKSEL	RX_CHAR
	MOVLW	'r'-1		;{'r','s','t'} as {AN0, AN1, AN2}
	SUBWF	RX_CHAR,W
	JBE	ADC_NEXT
	MOVLW	't'+1
	SUBWF	RX_CHAR,W
	JAE	ADC_NEXT

	;B1.ANSEL
	;bi: ANSi=0 (Digital), ANSi=1 (Analog)
	MOVLW	0x07
	BANKSEL	ANSEL
	MOVWF	ANSEL

	;B1.ADCON1
	;b7: unimp=0, b6..b4: ADCS2..ADCS0=0b010 (32Tosc)
	;b3..b0: unimp=0
	MOVLW	0x20
	BANKSEL	ADCON1
	MOVWF	ADCON1

	;Setup and start ADCON0
	BANKSEL	RX_CHAR
	MOVLW	'r'
	SUBWF	RX_CHAR,W
	MOVWF	ADCON0_TMP
	BCF	STATUS,C
	RLF	ADCON0_TMP,F
	RLF	ADCON0_TMP,W
	IORLW	ADCON0_M_ADFM | ADCON0_M_ADON
	MOVWF	ADCON0
	CALL	WAIT_5US
	CALL	WAIT_5US
	CALL	WAIT_5US
	CALL	WAIT_5US
	BSF	ADCON0,GO
	BTFSC	ADCON0,GO
	GOTO	$-1
	BANKSEL	ADRESH
	MOVF	ADRESH,W
	CALL	CONV_HEX
	MOVF	CONV_HEX_L,W
	BTFSS	PIR1,TXIF	;Wait TX ready.
	GOTO	$-1
	MOVWF	TXREG		;Echo back

	BANKSEL	ADRESL
	MOVF	ADRESL,W
	BANKSEL	CONV_HEX_H
	CALL	CONV_HEX
	MOVF	CONV_HEX_H,W
	BTFSS	PIR1,TXIF	;Wait TX ready.
	GOTO	$-1
	MOVWF	TXREG		;Echo back

	BANKSEL	CONV_HEX_L
	MOVF	CONV_HEX_L,W
	BANKSEL	PIR1
	BTFSS	PIR1,TXIF	;Wait TX ready.
	GOTO	$-1
	MOVWF	TXREG		;Echo back

ADC_NEXT
	BANKSEL	PIR1
	GOTO	MAIN_LOOP

WAIT_5US_PLUS
	NOP
WAIT_5US
	NOPGOTO	4,(INST_FREQ / D'200000')	;200KHz as 5us
	RETURN

;@param  W Convert to Hex.
;@result CONVHEX_L	Hex char Low 4bit
;@result CONVHEX_H	Hex char High 4bit
CONV_HEX
	MOVWF	CONV_HEX_H
	ANDLW	0xf
	MOVWF	CONV_HEX_L
	ADDLW	(0xF6)		;W>=10, makes C=1. W<10, makes C=0.
	MOVLW	A'0'		;C=0, add '0'
	BTFSC	STATUS,C
	MOVLW	(A'A'-D'10')	;C=1, add 'A'-10
	ADDWF	CONV_HEX_L,F
	SWAPF	CONV_HEX_H,W
	ANDLW	0xf
	MOVWF	CONV_HEX_H
	ADDLW	(0xF6)		;W>=10, makes C=1. W<10, makes C=0.
	MOVLW	A'0'		;C=0, add '0'
	BTFSC	STATUS,C
	MOVLW	(A'A'-D'10')	;C=1, add 'A'-10
	ADDWF	CONV_HEX_H,F
	RETURN

	END
