;PIC extend instructions.
;
;  Copyright 2013 Akinori Furuta<afuruta@m7.dion.ne.jp>.
;  All rights reserved.
;
;  Redistribution and use in source and binary forms, with or without 
;  modification, are permitted provided that the following conditions 
;  are met:
;
;  1. Redistributions of source code must retain the above copyright notice,
;     this list of conditions and the following disclaimer.
;
;  2. Redistributions in binary form must reproduce the above copyright notice,
;     this list of conditions and the following disclaimer in the documentation
;     and/or other materials provided with the distribution.
;
;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
;  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
;  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
;  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
;  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
;  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;Jump equal after SUBWF.
;JUMP."reg==W"	label
JE	MACRO	label
	BTFSC	STATUS,Z
	GOTO	label
	ENDM

;Jump below or equal, use after SUBWF.
;JUMP."reg<=W"	label
JBE	MACRO	label
	BTFSC	STATUS,Z
	GOTO	label		;Z=1
	BTFSS	STATUS,C
	GOTO	label		;C=0
	ENDM

;Jump below, use after SUBWF.
;JUMP."reg<W"	label
JB	MACRO	label
	LOCAL	j1
	BTFSC	STATUS,Z
	GOTO	j1		;Z=1
	BTFSS	STATUS,C
	GOTO	label		;C=0 SUBWF reg; reg<W
j1
	ENDM

;Jump above or equal, use after SUBWF.
;JUMP."reg>=W"	label
JAE	MACRO	label
	BTFSC	STATUS,C
	GOTO	label		;C=1 SUBWF reg; reg>=W
	ENDM

;Jump above, use after SUBWF.
;JUMP."reg>W"	label
JA	MACRO	label
	LOCAL	j1
	BTFSC	STATUS,Z
	GOTO	j1		;Z=1
	BTFSC	STATUS,C
	GOTO	label		;C=1
j1
	ENDM

;Do next instruction if Z=1
DOZ	MACRO
	BTFSC	STATUS,Z
	ENDM

;Do next instruction if Z=0
DONZ	MACRO
	BTFSS	STATUS,Z
	ENDM

;Do next instruction if C=1
DOC	MACRO
	BTFSC	STATUS,C
	ENDM

;Do next instruction if C=0
DONC	MACRO
	BTFSS	STATUS,C
	ENDM

;Do next instruction after SUBWF reg, if reg>=W (above or equal)
DOAE	MACRO
	BTFSC	STATUS,C
	ENDM

;Do next instruction after SUBWF reg, if reg>W (above)
DOA	MACRO
	BTFSC	STATUS,Z
	GOTO	$+3
	BTFSC	STATUS,C
	ENDM

;Do next instruction after SUBWF reg, if reg<W (below)
DOB	MACRO
	BTFSS	STATUS,C
	ENDM

;Do next instruction after SUBWF reg, if reg<=W (below)
DOBE	MACRO
	BTFSC	STATUS,Z
	GOTO	$+2
	BTFSS	STATUS,C
	ENDM

;Do next instruction if reg's sign bit=1
DOSF	MACRO	reg
	BTFSC	reg,7
	ENDM

;Do next instruction if reg's sign bit=0
DONSF	MACRO	reg
	BTFSC	reg,7
	ENDM


MOVXW	MACRO	rsrc,rdst
	MOVF	rsrc,W
	MOVWF	rdst
	ENDM

;MOVXW with banksel
MOVXWB	MACRO	rsrc,rdst
	BANKSEL	rsrc
	MOVF	rsrc,W
	BANKSEL	rdst
	MOVWF	rdst
	ENDM

;rsrc + rdst + CY -> rdst
ADCXW	MACRO	rsrc,rdst
	MOVF	rsrc,W
	BTFSC	STATUS,C
	INCFSZ	rsrc,W
	ADDWF	rdst,F	;if (w!=0) {rdst=rdst+w;}; This keeps CY to propagate next digit.
	ENDM

;rsrc + rdst + CY -> rdst
ADCXWB	MACRO	rsrc,rdst
	BANKSEL	rsrc
	MOVF	rsrc,W
	BTFSC	STATUS,C
	INCFSZ	rsrc,W
	BANKSEL	rdst
	BTFSS	STATUS,Z
	ADDWF	rdst,F
	ENDM

;rsrc + rdst  -> rdst
ADDXW	MACRO	rsrc,rdst
	MOVF	rsrc,W
	ADDWF	rdst,F
	ENDM

;rsrc + rdst -> rdst
ADDXWB	MACRO	rsrc,rdst
	BANKSEL	rsrc
	MOVF	rsrc,W
	BANKSEL	rdst
	ADDWF	rdst,F
	ENDM


;BO=/CY
;rdst - rsrc - BO -> rdst

SBBXW	MACRO	rsrc,rdst
	MOVF	rsrc,W
	BTFSS	STATUS,C
	INCFSZ	rsrc,W
	SUBWF	rdst,F
	ENDM

;SBB with banksel
SBBXWB	MACRO	rsrc,rdst
	BANKSEL	rsrc
	MOVF	rsrc,W
	BTFSS	STATUS,C
	INCFSZ	rsrc,W
	BANKSEL	rdst
	BTFSS	STATUS,Z
	SUBWF	rdst,F
	ENDM

;BO=/CY
;rdst - rsrc -> rdst

SUBXW	MACRO	rsrc,rdst
	MOVF	rsrc,W
	SUBWF	rdst,F
	ENDM

;SUB with banksel
SUBXWB	MACRO	rsrc,rdst
	BANKSEL	rsrc
	MOVF	rsrc,W
	BANKSEL	rdst
	SUBWF	rdst,F
	ENDM

XCHGWF	MACRO	reg
	XORWF	reg,F
	XORWF	reg,W
	XORWF	reg,F
	ENDM

;reg=-reg;
NEGF8	MACRO	reg
	COMF	reg,F
	INCF	reg,F
	ENDM

;reg=abs(reg);
ABSF	MACRO	reg
	BTFSS	reg,0x7
	GOTO	$+3
	COMF	reg,F
	INCF	reg,F
	ENDM

;@return rdst==0: positive(rsrc), rdst==0xff: negative(rsrc)
SGNE	MACRO	rsrc, rdst
	CLRF	rdst
	BTFSC	rsrc,0x7
	DECF	rdst
	ENDM

; reg=(reg==0 ? 0x00 : (reg<0 ? 0xff : 0x01))
SIGNF	MACRO	reg
	LOCAL	reg_zero
	MOVF	reg,F
	BTFSC	STATUS,Z
	GOTO	$+6		;0
	RLF	reg,F		;1
	CLRF	reg		;2
	BTFSC	STATUS,C	;3
	DECF	reg,F		;4
	BSF	reg,0x0		;5
	ENDM			;6

;unsigned compare.
;@return C==1: rsrc_h:rsrc_l >= const16, C==0: rsrc_h:rsrc_l < const16
;        Z==1: Equal, Z==0: Not Equal.
;
UCMP16L	MACRO	rsrc_h, rsrc_l, const16
	LOCAL	const16_l, const16_h, j_hne
const16_l = (const16 >> 0x8) & 0xff
const16_h = (const16 & 0xff)
	MOVLW	const16_h
	SUBWF	rsrc_h,W
	BTFSS	STATUS,Z
	GOTO	j_hne
	MOVLW	const16_l
	SUBWF	rsrc_l,W
j_hne
	ENDM

;unsigned compare.
;@return C==1: ra_h:ra_l >= rb_h:rb_l, C==0: ra_h:ra_l < rb_h:rb_l
;        Z==1: Equal, Z==0: Not Equal.
;
UCMP16F	MACRO	ra_h, ra_l, rb_h, rb_l
	LOCAL	j_hne
	MOVF	rb_h,W
	SUBWF	ra_h,W
	BTFSS	STATUS,Z
	GOTO	j_hne
	MOVF	rb_l,W
	SUBWF	ra_l,W
j_hne
	ENDM

NOPGOTO	MACRO	cyc_spent,cyc_max
	LOCAL	cyc_diff, i_goto_n, i_nop_n
cyc_diff=cyc_max-cyc_spent
	IF cyc_spent < cyc_max

i_goto_n=cyc_diff / D'2'
		WHILE	i_goto_n > 0
			GOTO	$+1
i_goto_n--
		ENDW

i_nop_n=cyc_diff % D'2'
		WHILE	i_nop_n > 0
			NOP
i_nop_n--
		ENDW
	ENDIF
	ENDM
